from setuptools.command.install import install
import os

class CustomInstall(install):
    def run(self):
        install.run(self)

        if not os.path.exists('models'):
            os.makedirs('models')

        if not os.path.exists('client_repo'):
            os.makedirs('client_repo')

        model_py_content = """
        import tensorflow as tf

        class MLMODEL:
            def __init__(self):
                self.model = tf.keras.models.Sequential() # write your model code
                self.model.compile() # write your model compile code

            def getModel(self):
                return self.model
        """

        with open('client_repo/MODEL.py', 'w') as f:
            f.write(model_py_content)


import tensorflow as tf
import numpy as np
import json
import boto3
import importlib.util
from code_repo import MODEL
import os


class FLClient(object):
    def __init__(self, member_ID, x_train_client, y_train_client, x_test_client, y_test_client):
        self.nodeId = member_ID

        self.FLTaskName = None
        self.roundId = -1
        self.serverId = -1
        self.epochs_client = -1

        self.countTransaction = 1
        self.num_train_per_client = 0
        self.num_test_per_client = 0

        self.x_train_client, self.y_train_client, self.x_test_client, self.y_test_client = x_train_client, y_train_client, x_test_client, y_test_client
        self.num_train_per_client = self.x_train_client.shape[0]
        self.num_test_per_client = self.x_test_client.shape[0]
        print("Local datasets:")
        print("x_train =", self.x_train_client.shape)
        print("y_train =", self.y_train_client.shape)
        print("x_test =", self.x_test_client.shape)
        print("y_test =", self.y_test_client.shape)
        print()

    def receiveNotificationsFromServer(self, sqs_region, sqs_name, cross_account_sqs_role=None):
        if cross_account_sqs_role == None:
            sqs_obj = boto3.resource('sqs', region_name=sqs_region)
            sqs_queue = sqs_obj.get_queue_by_name(QueueName=sqs_name)
            sqs_msgs = sqs_queue.receive_messages(
                AttributeNames=['All'],
                MessageAttributeNames=['All'],
                VisibilityTimeout=15,
                WaitTimeSeconds=20,
                MaxNumberOfMessages=5
            )

            sqs_client = boto3.client('sqs', region_name=sqs_region)
            queue_url = sqs_client.get_queue_url(QueueName=sqs_name)['QueueUrl']
            for msg in sqs_msgs:
                sqs_client.delete_message(QueueUrl=queue_url,
                                          ReceiptHandle=msg.receipt_handle)
            return sqs_msgs

        elif cross_account_sqs_role is not None:
            client = boto3.client('sts')
            response = client.assume_role(RoleArn=cross_account_sqs_role,
                                          RoleSessionName='access-cross-account-sqs',
                                          DurationSeconds=900)

            sqs_obj = boto3.resource('sqs', region_name=sqs_region,
                                     aws_access_key_id=response['Credentials']['AccessKeyId'],
                                     aws_secret_access_key=response['Credentials']['SecretAccessKey'],
                                     aws_session_token=response['Credentials']['SessionToken'])

            sqs_queue = sqs_obj.get_queue_by_name(QueueName=sqs_name)
            sqs_msgs = sqs_queue.receive_messages(
                AttributeNames=['All'],
                MessageAttributeNames=['All'],
                VisibilityTimeout=15,
                WaitTimeSeconds=20,
                MaxNumberOfMessages=5
            )

            sqs_client = boto3.client('sqs', region_name=sqs_region,
                                      aws_access_key_id=response['Credentials']['AccessKeyId'],
                                      aws_secret_access_key=response['Credentials']['SecretAccessKey'],
                                      aws_session_token=response['Credentials']['SessionToken'])
            queue_url = sqs_client.get_queue_url(QueueName=sqs_name)['QueueUrl']
            for msg in sqs_msgs:
                sqs_client.delete_message(QueueUrl=queue_url,
                                          ReceiptHandle=msg.receipt_handle)
            return sqs_msgs

    def localTraining(self, model_params_w_file, s3_fl_model_registry, cross_account_s3_role=None):

        if (self.roundId == 0):
            print("2: Download a global model")
            mlmodel = MODEL.MLMODEL()
            model = mlmodel.getModel()
            prio_train_loss, prio_train_acc = "NA", "NA"
            prio_test_loss, prio_test_acc = "NA", "NA"
            print("prio local training: training loss: {} \t training accuracy: {}".format(prio_train_loss,
                                                                                           prio_train_acc))
            print("prio local training: testing loss: {} \t testing accuracy: {}".format(prio_test_loss, prio_test_acc))
            print()

            print("3: Local training ...")
            model.fit(self.x_train_client, self.y_train_client, epochs=self.epochs_client)
            after_train_loss, after_train_acc = model.evaluate(self.x_train_client, self.y_train_client, verbose=2)
            after_test_loss, after_test_acc = model.evaluate(self.x_test_client, self.y_test_client, verbose=2)
            print('after local training: training loss: %2.4f \t training accuracy: %2.4f' % (
            after_train_loss, after_train_acc))
            print('after local training: testing loss: %2.4f \t testing accuracy: %2.4f \n' % (
            after_test_loss, after_test_acc))

        else:
            mlmodel = MODEL.MLMODEL()
            model = mlmodel.getModel()

            print("2: Download a global model")

            s3 = None
            if cross_account_s3_role == None:
                s3 = boto3.resource('s3')
            elif cross_account_s3_role is not None:
                client = boto3.client('sts')
                response = client.assume_role(RoleArn=cross_account_s3_role,
                                              RoleSessionName='access-cross-account-s3',
                                              DurationSeconds=900)
                s3 = boto3.resource(
                    "s3",
                    aws_access_key_id=response['Credentials']['AccessKeyId'],
                    aws_secret_access_key=response['Credentials']['SecretAccessKey'],
                    aws_session_token=response['Credentials']['SessionToken']
                )

            s3.Bucket(s3_fl_model_registry).download_file(model_params_w_file, "models/" + model_params_w_file)
            weights_prio_training = np.load("models/" + model_params_w_file, allow_pickle=True)
            model.set_weights(weights_prio_training)

            prio_train_loss, prio_train_acc = model.evaluate(self.x_train_client, self.y_train_client, verbose=2)
            prio_test_loss, prio_test_acc = model.evaluate(self.x_test_client, self.y_test_client, verbose=2)
            print('prio local training: training loss: %2.4f \t training accuracy: %2.4f' % (
            prio_train_loss, prio_train_acc))
            print(
                'prio local training: testing loss: %2.4f \t testing accuracy: %2.4f' % (prio_test_loss, prio_test_acc))
            print()

            print("3: Local training ...")

            model.fit(self.x_train_client, self.y_train_client, epochs=self.epochs_client)
            after_train_loss, after_train_acc = model.evaluate(self.x_train_client, self.y_train_client, verbose=2)
            after_test_loss, after_test_acc = model.evaluate(self.x_test_client, self.y_test_client, verbose=2)
            print('after local training: training loss: %2.4f \t training accuracy: %2.4f' % (
            after_train_loss, after_train_acc))
            print('after local training: testing loss: %2.4f \t testing accuracy: %2.4f \n' % (
            after_test_loss, after_test_acc))

        savedModelFileName = 'train_weight_round_{}_client_{}.npy'.format(self.roundId, self.nodeId)
        weights = model.get_weights()
        np.save("models/" + savedModelFileName, weights)

        metricDict = {
            "Task_Name": self.FLTaskName,
            "Task_ID": str(self.nodeId).zfill(4) + str(self.countTransaction).zfill(8),
            "roundId": str(self.roundId),
            "member_ID": str(self.nodeId),
            "numSamples": str(self.num_train_per_client),
            "numClientEpochs": str(self.epochs_client),
            "trainLoss": str(after_train_loss),
            "trainAcc": str(after_train_acc),
            "testLoss": str(after_test_loss),
            "testAcc": str(after_test_acc),
            "weightsFile": savedModelFileName,
            "numClientsRequired": "NA",
            "source": "Client",
        }

        return savedModelFileName, metricDict

    def processGlobalModelInfoFromServer(self, transactions):

        tasks = set()

        for task in transactions:
            tasks.add(task["Task_Name"])

        latestTaskName = None
        if len(tasks) == 1:
            latestTaskName = tasks.pop()
        else:
            for taskName in list(tasks):
                if latestTaskName is None or latestTaskName < taskName:
                    latestTaskName = taskName

        latestRound = -1
        for transaction in transactions:
            if latestTaskName == transaction["Task_Name"] and latestRound < int(transaction["roundId"]):
                latestRound = int(transaction["roundId"])

        if self.FLTaskName == None or (self.FLTaskName <= latestTaskName and self.roundId < latestRound):
            self.FLTaskName = latestTaskName
            self.roundId = latestRound

        elif self.FLTaskName > latestTaskName:
            pass

        for transaction in transactions:
            if self.FLTaskName == transaction["Task_Name"] and self.roundId == int(transaction["roundId"]):
                metricDict = {
                    "roundId": transaction["roundId"],
                    "member_ID": transaction["member_ID"],
                    "numSamples": transaction["numSamples"],
                    "numClientEpochs": transaction["numClientEpochs"],
                    "trainAcc": float(transaction["trainAcc"]) if transaction["trainAcc"] != "NA" else "NA",
                    "testAcc": float(transaction["testAcc"]) if transaction["testAcc"] != "NA" else "NA",
                    "trainLoss": float(transaction["trainLoss"]) if transaction["trainLoss"] != "NA" else "NA",
                    "testLoss": float(transaction["testLoss"]) if transaction["testLoss"] != "NA" else "NA",
                    "weightsFile": transaction["weightsFile"],
                    "numClientsRequired": transaction["numClientsRequired"],
                    "source": transaction["source"],
                    "TaskToken": transaction["TaskToken"],
                }
                self.epochs_client = int(metricDict["numClientEpochs"])
                self.serverId = metricDict["member_ID"]
                return metricDict
        return None

    def uploadToFLServer(self, s3_fl_model_registry, savedModelFileName, dynamodb_table_model_info, metricDict,
                         cross_account_s3_role=None, cross_account_dynamodb_role=None):
        s3 = None
        dynamodb = None
        if cross_account_s3_role == None and cross_account_dynamodb_role == None:
            s3 = boto3.resource('s3')
            dynamodb = boto3.resource('dynamodb')
        elif cross_account_s3_role is not None and cross_account_dynamodb_role is not None:
            client = boto3.client('sts')
            response = client.assume_role(RoleArn=cross_account_s3_role,
                                          RoleSessionName='access-cross-account-s3',
                                          DurationSeconds=900)
            s3 = boto3.resource(
                "s3",
                aws_access_key_id=response['Credentials']['AccessKeyId'],
                aws_secret_access_key=response['Credentials']['SecretAccessKey'],
                aws_session_token=response['Credentials']['SessionToken']
            )

            dynamodb_client = boto3.client('sts')
            dynamodb_response = dynamodb_client.assume_role(RoleArn=cross_account_dynamodb_role,
                                                            RoleSessionName='access-cross-account-dynamoDB',
                                                            DurationSeconds=900)
            dynamodb = boto3.resource('dynamodb',
                                      aws_access_key_id=dynamodb_response['Credentials']['AccessKeyId'],
                                      aws_secret_access_key=dynamodb_response['Credentials']['SecretAccessKey'],
                                      aws_session_token=dynamodb_response['Credentials']['SessionToken'])

        s3.Bucket(s3_fl_model_registry).upload_file("models/" + savedModelFileName, savedModelFileName)

        task_table = dynamodb.Table(dynamodb_table_model_info)
        task_name = metricDict["Task_Name"]
        task_ID = metricDict["Task_ID"]
        task_table.put_item(Item=metricDict)

        printDiction = {
            "taskName": metricDict["Task_Name"],
            "taskId": metricDict["Task_ID"],
            "roundId": metricDict["roundId"],
            "memberId": metricDict["member_ID"],
            "numSamples": metricDict["numSamples"],
            "trainAcc": float(metricDict["trainAcc"]) if metricDict["trainAcc"] != "NA" else "NA",
            "testAcc": float(metricDict["testAcc"]) if metricDict["testAcc"] != "NA" else "NA",
            "trainLoss": float(metricDict["trainLoss"]) if metricDict["trainLoss"] != "NA" else "NA",
            "testLoss": float(metricDict["testLoss"]) if metricDict["testLoss"] != "NA" else "NA",
            "weightsFile": metricDict["weightsFile"],
            "source": metricDict["source"],
            "taskToken": metricDict["TaskToken"],
        }
        print("Upload local model information: {} \n\n".format(str(printDiction)))
        self.countTransaction = self.countTransaction + 1

    def downloadFLGlobalModel(self, s3_fl_model_registry, cross_account_s3_role=None):
        print("6: Download the FL trained global model for deployment")

        s3 = None
        if cross_account_s3_role == None:
            s3 = boto3.resource('s3')
        elif cross_account_s3_role is not None:
            client = boto3.client('sts')
            response = client.assume_role(RoleArn=cross_account_s3_role,
                                          RoleSessionName='access-cross-account-s3',
                                          DurationSeconds=900)
            s3 = boto3.resource(
                "s3",
                aws_access_key_id=response['Credentials']['AccessKeyId'],
                aws_secret_access_key=response['Credentials']['SecretAccessKey'],
                aws_session_token=response['Credentials']['SessionToken']
            )

        model_params_w_file = "train_weight_round_" + str(self.roundId - 1) + ".npy"
        print(model_params_w_file)
        s3.Bucket(s3_fl_model_registry).download_file(model_params_w_file, "models/" + model_params_w_file)
        return model_params_w_file



import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler, LabelEncoder


def input_fn(file_path, target_col):
    train_dataframe = pd.read_csv(file_path)

    x_data = train_dataframe.drop(target_col, axis=1)
    y_data = train_dataframe[target_col]
    pre_x_train = data_preprocess(x_data)

    x_train_client, x_test_client, y_train_client, y_test_client = train_test_split(pre_x_train, y_data)

    return x_train_client, y_train_client, x_test_client, y_test_client


def data_preprocess(dataframe):
    label_encoder = LabelEncoder()
    for column in dataframe.columns:
        if dataframe[column].dtype == 'object':
            dataframe[column] = label_encoder.fit_transform(dataframe[column])

    scaler = StandardScaler()
    numerical_data = dataframe.select_dtypes(include=['int64', 'float64'])
    scaled_numerical_data = scaler.fit_transform(numerical_data)
    dataframe[numerical_data.columns] = scaled_numerical_data

    return dataframe


import sys
import boto3
from botocore.exceptions import NoCredentialsError, ClientError

import numpy as np
import tensorflow as tf # use tensorflow for local training
import json
import warnings
from .client_fedlearn import FLClient
import time
import os
import glob
import shutil
warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning) 


def FL_start(member_ID, x_train_client, y_train_client, x_test_client, y_test_client):    

    start_time = 0
    end_time = 0
    hasStarted = True

    config_client = './client_config.json'
    global_model = None
    
    with open(config_client, 'r') as config_file:
        client_config = json.load(config_file)
    s3_fl_model_registry = client_config["s3_fl_model_registry"]
    dynamodb_table_model_info = client_config["dynamodb_table_model_info"]
    member_ID = int(client_config["clients"][member_ID-1]["member_ID"])
    sqs_region = client_config["clients"][member_ID-1]["sqs_region"]
    client_queue_name = client_config["clients"][member_ID-1]["client_queue_name"]
    
    cross_account = False
    if (
        "cross_account_sqs_role" in client_config.keys() and 
        "cross_account_s3_role" in client_config.keys() and
        "cross_account_dynamodb_role" in client_config.keys() 
       ):
        cross_account = True
    else:
        assert("cross_account_sqs_role" not in client_config.keys() and 
               "cross_account_s3_role" not in client_config.keys() and
               "cross_account_dynamodb_role" not in client_config.keys())
    
    client = FLClient(member_ID, x_train_client, y_train_client, x_test_client, y_test_client)
    
    signalTerminate = False
    while not signalTerminate:
        if cross_account == True:
            messages = client.receiveNotificationsFromServer(sqs_region, client_queue_name, client_config["cross_account_sqs_role"])
        else:
            messages = client.receiveNotificationsFromServer(sqs_region, client_queue_name)
    
        transactions = []
        curr_round = -1
        for msg in messages:
            msg_body = msg.body
            msg_body_json = json.loads(msg_body)
            msg_rec = msg_body_json["Message"]
            transaction_dict = json.loads(msg_rec)  
    
            if type(transaction_dict) == str:
                transaction_dict = json.loads(transaction_dict)
                if transaction_dict['roundId'] == "NA":
                    if cross_account == True:
                        global_model = client.downloadFLGlobalModel(s3_fl_model_registry, client_config["cross_account_s3_role"])
                    else: 
                        global_model = client.downloadFLGlobalModel(s3_fl_model_registry)
                    end_time = time.monotonic()
                    print("FL training is finished")
                    signalTerminate = True
                    break             
            
            transaction = transaction_dict["Input"]
            transaction["TaskToken"] = transaction_dict["TaskToken"]
            transactions.append(transaction)
            
            if int(transaction['roundId']) > curr_round:
                curr_round = transaction['roundId']
            
        if len(transactions) > 0:
            s3_client = boto3.client('s3')
            file_path = 'client_repo/MODEL.py'
            bucket_name = 'b3o'
            s3_file_path = f'fedlearn/model/{transaction["Task_Name"]}-model.py'

            try:
                s3_client.head_object(Bucket=bucket_name, Key=s3_file_path)
            except ClientError as e:
                error_code = int(e.response['Error']['Code'])
                if error_code == 404:
                    s3_client.upload_file(file_path, bucket_name, s3_file_path)
                else:
                    print(f"Error occurred: {e}")


            infoServer = client.processGlobalModelInfoFromServer(transactions)

            if (infoServer != None):
                if hasStarted:
                    start_time = time.monotonic()
                    hasStarted = False
                print("FL training round: " + str(curr_round) + "\n")
                print("1: Receive notification")            
                printDiction = {
                        "taskName": transaction["Task_Name"],
                        "taskId": transaction["Task_ID"],
                        "roundId":  transaction["roundId"],
                        "memberId": transaction["member_ID"],
                        "numClientEpochs": transaction["numClientEpochs"],
                        "trainAcc": float(transaction["trainAcc"]) if transaction["trainAcc"] != "NA" else "NA", 
                        "testAcc" : float(transaction["testAcc"]) if transaction["testAcc"] != "NA" else "NA", 
                        "trainLoss": float(transaction["trainLoss"]) if transaction["trainLoss"] != "NA" else "NA", 
                        "testLoss" : float(transaction["testLoss"]) if transaction["testLoss"] != "NA" else "NA", 
                        "weightsFile": transaction["weightsFile"],
                        "numClientsRequired": transaction["numClientsRequired"],
                        "source": transaction["source"],         
                        "taskToken": transaction["TaskToken"],
                    }                        
                print("Received notification: {} \n".format(str(printDiction))) 
                
                global_model_name = infoServer["weightsFile"]
            
                if cross_account == True:
                    local_model_name, local_model_info  = client.localTraining(global_model_name, s3_fl_model_registry, client_config["cross_account_s3_role"])
                else:
                    local_model_name, local_model_info = client.localTraining(global_model_name, s3_fl_model_registry)
                
                taskToken = infoServer["TaskToken"]
                local_model_info["TaskToken"] = taskToken
    
                print("4: Upload local model & its information ...")
                if cross_account == True:
                    client.uploadToFLServer(s3_fl_model_registry, 
                                            local_model_name, 
                                            dynamodb_table_model_info, 
                                            local_model_info, 
                                            client_config["cross_account_s3_role"], 
                                            client_config["cross_account_dynamodb_role"]
                                           )
                else:
                    client.uploadToFLServer(s3_fl_model_registry, 
                                            local_model_name, 
                                            dynamodb_table_model_info,
                                            local_model_info)
    
                client.roundId = client.roundId + 1

    
    if signalTerminate:   
        
        model_folder_path = './models'
        bucket_name = 'b3o'
        s3_folder_path = 'fedlearn/output/'
        task_name = transaction["Task_Name"]
        upload_file_name = f'{task_name}-output.npy'

        
        files = glob.glob(f'{model_folder_path}/*')
        latest_file = max(files, key=os.path.getmtime)
    
        
        s3 = boto3.client('s3')
        s3.upload_file(latest_file, bucket_name, s3_folder_path + upload_file_name)
            
        client_folder_path = './client_repo'
        for item in glob.glob(client_folder_path + '/*'):
            if os.path.isfile(item):
                os.remove(item)

        for item in glob.glob(model_folder_path + '/*'):
            if os.path.isfile(item):
                os.remove(item)

  




import setuptools
import sys

sys.path.append('src')
from b3o_fedlearn.client_setup import CustomInstall


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="b3o-fedlearn",
    version="0.0.10",
    author="skswjdekdud",
    author_email="skswjdekdud1103@gmail.com",
    description="a project for federated learning using AWS",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ACS-High-School",
    project_urls={
        "Bug Tracker": "https://github.com/ACS-High-School/ML/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    cmdclass={
        'install': CustomInstall,
    },
    python_requires=">=3.8",
)
